

//	
//																  
//			PictInfo -- Omnis 7 extended commands				  
//																  
//			by Kelly Burgess				June 3, 1998		  
//																  
//			CodeWarrior Pro 3 source code : Monaco 9, Tabs 4	  
//																  
//	

//
//	 PictInfo 7CMD resource -- copyright 1996 by Kelly Burgess
//
//	  An extended command for getting the size of a picture field
//	  An extended command for comparing two picture fields
//
//	 This external code resource / fragment / Win32 DLL is an extended commands type:
//	  The commands defined here will appear in the Omnis 7 commands list
//	  They can be invoked like any other procedure command
//
//
//	--------------------------------------------------------
//
//		!! critical that PictInfo is first in link order !!
//
//		The only modification needed to convert MPW source to CodeWarrior
//		was renaming the main entry point as 'main' instead of 'DoPictInfo'
//	--------------------------------------------------------

//  syntax  
//
//	 It's not necessary to load this external before calling it
//	 It's not possible to call this external either!
//
//	This creates two extended commands, which will appear in the commands list in Omnis 7 v1.1 and up.
//	One command can be invoked to obtain the size of a picture field.
//	The other command can be invoked to compare two picture fields.
//	The syntax is:
//	  GetPictSize(pictureField) with return value
//	   e.g. GetPictSize(MY_PICT) with return value #1
//		#1 will be set to 0 or greater if no errors occur; the value is the size in bytes
//		#1 will be set to a negative value if errors occur
//
//		 1..n = no errors, field is [1..n] bytes in size
//		 0 = no errors, field is empty
//		-1 = field not found
//		-2 = bad parameters
//		-3 = insufficient memory
//
//	  HasPictChanged(firstPictureField, secondPictureField) with return value
//	   e.g. HasPictChanged(BEFORE_PICT,AFTER_PICT) with return value #1
//		#1 will be set to 0 if no errors occur and the pictures are equal,
//		#1 will be set to 1 if no errors occur and the pictures are not equal,
//		#1 will be set to a negative value for error conditions...
//
//		 0 = no errors, pictures are equal
//		 1 = no errors, pictures don't match
//		-1 = field not found
//		-2 = bad parameters
//		-3 = insufficient memory
//
//		( Win32 note: picture comparison is not fully implemented,
//					  but is "left as an exercise for the programmer" )
//
//  modification history  
//
//	Tue, Nov 12, 1996 - started sample shell extended commands...
//	Sat, Jun  7, 1998 - made it cross-platform with multiple targets for CWPro
//	Sat, Jun  7, 1998 - note that for Win32, HasPictChanged is NOT fully implemented!
//	Wed, Jun 24, 1998 - Win32 bug fix - was calling GlobalFree() on an object handle; should use DeleteObject()
//
//  includes  

#include	"Callback.h"			//	Omnis callbacks (defines MACOS or WIN32OS)

#if defined(MACOS)
 #include	<string.h>				//	Macintosh include files
 #ifdef powerc
  #include	<fp.h>					//	latest Callback.h 17-Nov-97 already has this
 #endif

#elif defined (WIN32OS)
 #include	<string.h>

 #define	nil			0L
 #define	noErr		0

 #define	Boolean		BOOL

#else
 #error "No platform has been defined."
#endif


//  constants  

#define	ROOTID	30201		//	resource ID base

enum {						//	command selectors
	SIZE_REQUEST	= 1,		//	PictSize() command
	COMPARE_REQUEST	= 2			//	PictChanged() command
};

enum {						//	command return values
	MATCH			=  0,		//	PictChanged() fields match
	NOMATCH			=  1,		//	PictChanged() fields don't match
	FLDNOTFND		= -1,		//	a field wasn't found (unused)
	BADPARMS		= -2,		//	not enough parameters were passed
	MEMORYERR		= -3		//	there's not enough memory
};


//  WIN32 Globals  

//	globals:
#if defined(WIN32OS)	/* Win32 DLL Globals */

HINSTANCE	hInstLib;			//	DLL instance handle

#endif

//  prototypes  

//	main entry point:

#if defined(MACOS)
 #if defined(powerc)
  #pragma export on
   pascal void main(short md, long *ref, oproc callback);
  #pragma export off
 #else
  pascal void main(short md, long *ref, oproc callback);
 #endif

#elif defined(WIN32OS)
 BOOL WINAPI DllMain( HINSTANCE hInst, DWORD reason, LPVOID reserved);
 VOID FAR PASCAL PictInfo(short md, long far *ref, HWND hwnd, HWND topwindow, HINSTANCE instance, oproc callback);
#endif

Boolean IsPictureField( short fldRefNum, oproc callback);


	//	--------------  code starts here  --------------
	//	this is the first byte of the 'PictInfo' extended commands 7CMD resource
	//	this file must be linked first, if additional files are linked.

//	  PictInfo  
//	PROCEDURE PictInfo(md:integer; VAR ref:longint; callback:ptr);
//	

#if defined(MACOS)
pascal void main(short md, long *ref, oproc callback)

#elif defined(WIN32OS)
VOID FAR PASCAL PictInfo(short md, long far *ref, HWND hwnd, HWND topwindow, HINSTANCE instance, oproc callback)
#endif
{

#pragma unused (ref)
#if defined(WIN32OS)
#pragma unused (hwnd,topwindow,instance)
#endif

	short				errCd;				//	word, error code
	short				paramCount;			//	word, number of parameters
	short				cmdNumber;			//	word, number of command
	odouble				xtnded;				//	10-byte extended, scratch
	_extdesigninforec	*designinfo;		//	pointer to design info record
	long 				setupinfo;			//	long, pointer to setup or design info returned by GetDesignPtr
	long				longCount,byteCount; //	long, counters
	long				x;					//	long, for counting & indexing
	unsigned long		ph1Size;			//	long, size of picture
	unsigned long		ph2Size;			//	long, size of second picture
	unsigned long		*lp1, *lp2;			//	pointers to longs for scanning
	unsigned char		*cp1, *cp2;			//	pointers to chars for scanning
	char				strvalue[256];		//	to hold command/param string

   #if defined(MACOS)
	Handle				ph1;				//	handle for Mac picture
	Handle				ph2;				//	handle for second picture

   #elif defined(WIN32OS)
	HANDLE				ph1;				//	handle for Win32 BITMAP object
	HANDLE				ph2;				//	handle for second picture
	char				*pt1;				//	pointer for general use
	char				*pt2;				//	pointer for general use
	short				y;					//	word for counting & indexing
	short				colorBits;			//	bit depth of bitmap
	BITMAP				bmp, bmp2;			//	bitmap structures
   #endif

			//	this is one way to embed copyright info in code resource..
			//	be sure to use the -b compiler switch to keep strings with code
			//	also referred to as PC-relative strings in some environments

 	char				*copyright	=	"PictInfo Commands v1.01, Nov 12, 1996  1996 by Kelly Burgess";


			//	---------------- initialize locals -------------

	errCd	= 0;						//	no errors so far
	ph1		= nil;						//	clear handles
	ph2		= nil;

			//	--------------  case on md selector  --------------

	switch (md)
	{

	//  load / unload  

	case ext_load:											//	Omnis loading 7CMD automatically on startup
	case ext_unload:										//	Omnis unloading 7CMD automatically on quitting

			//	--------------  nothing to do in this example  --------------

		//	You can allocate memory on ext_load, store the handle or pointer in *ref,
		//	get it from *ref and use it on ext_call, and dispose it on ext_unload.

		//	(NEVER call GetPackageId() from Load or Unload tasks)
		//	(It's meaningless, and it crashes some versions of Omnis)

		return;												//	do nothing


	//  design info  

		//	here you could optionally make use of the designMode() function in Design.c
		//	 and the GetResourceStr() function in ExtUtils.c

	case ext_designinfo:												//	Omnis calling at design time..
																		//	a command is being programmed,
		cmdNumber = GetPackageId(callback);								//	get command ID
		designinfo = (_extdesigninforec *) GetDesignPtr(callback);		//	get pointer to design info record

		switch(designinfo->qfldid) {									//	see what Omnis wants...

			case CalcLabel:												//	Omnis wants a prompt for calculation entry field

				strcpy(designinfo->lable,(const char*)"\pParms: ");			//	start with 'Parms: ' (pascal string needed)

			 #if defined(MACOS)
				GetIndString((unsigned char*)strvalue,ROOTID,cmdNumber);	//	get command string from Mac STR# resource

			 #elif defined(WIN32OS)
				y = LoadString(hInstLib,ROOTID+cmdNumber-1,(LPSTR)strvalue,255);
				pt1 = (char*)(strvalue + y);								// get cmd string from Windoze STRINGTABLE rsrc.
				pt2 = pt1--;
				for (x=y; x>0; *pt2-- = *pt1--, x--) ;						//	convert to pstring
				*pt2 = y;
			 #endif

					//	The string is something like "GetPictSize(pictureField):30201"
					//	We want to supply the part between parentheses as the calculation field prompt..

				for (x=1 ; x<strvalue[0] && strvalue[x]!=':' ; x++ )	{
					;														//	find the colon
				}
				strvalue[x] = 0x00;
				strvalue[0] = x;											//	remove :xxxx from end
				for (x=1 ; x<strvalue[0] && strvalue[x]!=0x28 ; x++ )	{	//	walk up to the opening parenthesis
					;														//	remove command name from beginning
				}
				strcat (&designinfo->lable[1],&strvalue[x]);
				designinfo->lable[0] += strvalue[0]-x;						//	append '(parameters)' string
				designinfo->usedefault = false;
				break;


			case ReturnLabel:											//	Omnis wants a prompt for return code entry field

				strcpy(designinfo->lable,(const char*)"\pReturn Code :");	//	hand back a pascal string
				designinfo->usedefault = false;
				break;
		}
		break;


	//  design setup  

		//	here you could optionally make use of the designMode() function in Design.c

	case ext_designsetup:									//	Omnis calling at design time..
															//	a command is being programmed,
		setupinfo = GetDesignPtr(callback);					//	get pointer to design setup record

		//SET_CUST_BUTTON_VISIBLE_OFF(setupinfo);			//	don't show the custom button (no longer supported in Omnis 7v3)
		SET_RET_FIELD_VISIBLE_ON(setupinfo);				//	do show the return value field (see Callback.h)
		break;


	//  call response  
	case ext_call:

		//	Delay(3,&longCount);		//	MPW debug door - uncomment and tell Macsbug "atb Delay"...

			//	--------------  get paramcount, command selector  --------------

		GetFldVal(ref_parmcnt, fmt_double, 10, &xtnded, callback);
		paramCount = (short)xtnded;					//	get number of parameters
		cmdNumber = GetPackageId(callback);			//	get command ID


		switch (cmdNumber)							//	do the command
		{

		//  PictSize() command  

		case SIZE_REQUEST:			//	GetPictSize(pictureField) with return value pictureSize

				//	--------------  check for a parameter   --------------

			if (paramCount == 0)	{								//	check for at least 1 parameter
				errCd = BADPARMS;									//	if not enough parameters, return an error <--
				break;
			}
				//	--------------  make sure it's a picture field   --------------

			if (!IsPictureField(ref_parm1,callback))	{
				errCd = BADPARMS;									//	if not a picture field, return an error <--
				break;
			}

				//	--------------  get picture field size   --------------
				//	It seems that passing a handle to GetFldVal(...,fmt_hpict,..) when the picture field is empty
				//	will destroy the handle passed in, instead of setting it to empty..
				//	So we do a size check first with fmt_hbinlen, and skip GetFldVal(..fmt_hpict..) if size is zero.
				//	fmt_hbinlen works with picture fields as well as binary fields, but (on Mac) it appears to return
				//	a length 16 greater than the size of the picture handle returned by GetFldVal(..fmt_hpict.)
				//	If the field is empty, though, it does return 0, not 16.

			ph1Size = 0;
			GetFldVal(ref_parm1, fmt_hbinlen, 4, &ph1Size, callback);
			if (ph1Size > 0)	{

					//	--------------  get the picture   --------------
					//	Mac Omnis will resize and use an existing handle passed to GetFldVal(), if possible.
					//	However, the handle returned may not be the same as the one passed in...
					//	Win32 Omnis always disposes any handle passed in, and creates a new handle
					//	The handle returned by GetFldVal is yours, you're responsible for disposing it.
					
				ph1 = nil;
				GetFldVal(ref_parm1, fmt_hpict, ph1Size, &ph1, callback);	//	NOTE: maxSize param is ignored for handle calls

				if (ph1 != nil)	{									//	ph1 is the handle to the picture field data

				   #if defined(MACOS)
					ph1Size = GetHandleSize(ph1);					//	for Mac, just the handle size will do it

				   #elif defined(WIN32OS)
					//ph1Size = GlobalSize(ph1);					//	for Win32, this won't work - the handle is an object!
					ph2Size = GetObject(ph1,sizeof(BITMAP),&bmp);	//	get bitmap from handle, size returned is sizeof(BITMAP)
					colorBits = (short)(bmp.bmPlanes * bmp.bmBitsPixel);
					if		(colorBits == 1)	colorBits = 1;		//	get 'rounded' color depth
					else if (colorBits <= 4)	colorBits = 4;
					else if (colorBits <= 8)	colorBits = 8;
					else if (colorBits <= 16)	colorBits = 16;
					else if (colorBits <= 24)	colorBits = 24;
					else						colorBits = 32;

						//	depending on what you want, maybe you'd rather return the compressed size
						//	 which was obtained in the GetFldVal(..fmt_hbinlen..) call above...
					ph1Size = (bmp.bmWidth + 7)/8 * bmp.bmHeight * colorBits;	//	compute actual image size
				   #endif
				}
				else {
					ph1Size = 0;
				}
			}

				//	--------------  return the command result   --------------
				//	return size of pict in command's return value..

			xtnded = (odouble)ph1Size;
		   #if defined(powerc)
			//SetFldVal(ref_returnval, fmt_double, &xtnded, callback);	//	PPC version not working for this callback
			//SetFldLong(ref_returnval,ph1Size,callback);				//	or this one... (v3.6.4 callback file should fix this)
			NumToString(ph1Size,(unsigned char*)&strvalue[0]);			//	Meanwhile, convert to string and pass that back
			SetFldVal(ref_returnval, fmt_pstring, &strvalue[0], callback);	//	(problem is w/Metrowerks' support of long doubles)
		   #else
			SetFldVal(ref_returnval, fmt_double, &xtnded, callback);
		   #endif

			break;		//	end case SIZE_REQUEST


		//  PictChanged() command  

		case COMPARE_REQUEST:		//	HasPictChanged(firstPictureField, secondPictureField) with return value

				//	--------------  check for parameters   --------------

			if (paramCount < 2)	{									//	check for at least 2 parameters
				errCd = BADPARMS;									//	if not enough parameters, return an error <--
				break;
			}
				//	--------------  make sure they're both picture fields   --------------

			if (!IsPictureField(ref_parm1,callback))	{
				errCd = BADPARMS;									//	if parm 1 isn't a picture field, return an error <--
				break;
			}
			if (!IsPictureField(ref_parm2,callback))	{
				errCd = BADPARMS;									//	if parm 2 isn't a picture field, return an error <--
				break;
			}

				//	--------------  get first picture field size   --------------

			ph1 = nil;
			ph1Size	= 0;
			GetFldVal(ref_parm1, fmt_hbinlen, 4, &ph1Size, callback);
			if (ph1Size > 0)	{

					//	--------------  get first picture   --------------

				GetFldVal(ref_parm1, fmt_hpict, 0, &ph1, callback);
				if (ph1 != nil)	{
				   #if defined(MACOS)
					ph1Size = GetHandleSize(ph1);					//	get the size to compare first
				   #elif defined(WIN32OS)
					ph1Size = GetObject(ph1,sizeof(BITMAP),&bmp);	//	get bitmap from handle, size returned is sizeof(BITMAP)
				   #endif
				}
				else {
					ph1Size = 0;
				}
			}
				//	--------------  get second picture field size   --------------

			ph2 = nil;
			ph2Size = 0;
			GetFldVal(ref_parm2, fmt_hbinlen, 4, &ph2Size, callback);
			if (ph2Size > 0)	{

					//	--------------  get second picture   --------------

				GetFldVal(ref_parm2, fmt_hpict, 0, &ph2, callback);
				if (ph2 != nil)	{
				   #if defined(MACOS)
					ph2Size = GetHandleSize(ph2);					//	get the size to compare first
				   #elif defined(WIN32OS)
					ph2Size = GetObject(ph2,sizeof(BITMAP),&bmp2);	//	get bitmap from handle, size returned is sizeof(BITMAP)
				   #endif
				}
				else {
					ph2Size = 0;
				}
			}

				//	--------------  compare two pictures   --------------

			if (ph1Size == ph2Size)	{				//	if handle sizes match, scan the handle contents...

			   #if defined(MACOS)
				HLock(ph1);							//	lock the handles
				lp1 = (unsigned long*)*ph1;			//	see also: BLYSTD.H or PORTMAC.H for useful x-platform memory macros
				HLock(ph2);
				lp2 = (unsigned long*)*ph2;

			   #elif defined(WIN32OS)
				lp1 = (unsigned long*)&bmp;
				lp2 = (unsigned long*)&bmp2;
				ph1Size -= sizeof(bmp.bmBits);		//	don't compare final field, it's OK if bits pointers are different
				ph2Size -= sizeof(bmp.bmBits);		//	 (in fact they're both NIL at this point)
			   #endif

				longCount = ph1Size >> 2;			//	get number of longs in handles
				byteCount = ph1Size & 0x00000003;	//	get number of leftover bytes
				xtnded = MATCH;						//	be optimistic

				while (longCount-- > 0)	{			//	scan handles 4 bytes at a time
					if (*lp1++ != *lp2++)	{
						xtnded = NOMATCH;
						break;
					}
				}
				if (xtnded == MATCH)	{
					cp1 = (unsigned char*)lp1;
					cp2 = (unsigned char*)lp2;
					while (byteCount-- > 0)	{		//	scan leftover bytes one at a time
						if (*cp1++ != *cp2++)	{
							xtnded = NOMATCH;
							break;
						}
					}
				}
			   #if defined(MACOS)
				HUnlock(ph1);
				HUnlock(ph2);						//	unlock the handles
			   #elif defined(WIN32OS)
				if (xtnded == MATCH)	{			//	for windows, if basic bitmaps match,
					//	!!! NOTE !!!
					//	we still should get the actual bits and compare content...
					//	we know dimensions/depth etc. are the same, but that's all we know..
					//	obtaining actual bits and comparing them is "left as an exercise for the programmer"
				}
			   #endif
			}
			else {
				xtnded = NOMATCH;					//	skip scan when sizes differ
			}

				//	--------------  return the command result   --------------
				//	return MATCH or NOMATCH in command's return value

		   #if defined(powerc)
			//SetFldVal(ref_returnval, fmt_double, &xtnded, callback);	//	PPC version not working for this callback
			//SetFldLong(ref_returnval,ph1Size,callback);				//	or this one... (v3.6.4 callback file should fix it)
			NumToString((long)xtnded,(unsigned char*)&strvalue[0]);		//	Meanwhile, convert to string and pass that back
			SetFldVal(ref_returnval, fmt_pstring, &strvalue[0], callback);	//	(problem is w/Metrowerks' support of long doubles)
		   #else
			SetFldVal(ref_returnval, fmt_double, &xtnded, callback);
		   #endif

			break;		//	end case COMPARE_REQUEST

		}	//	end switch (cmdNumber)

		//	end case ext_call

	}	//	end switch (md)

	//  cleanup  

		//	--------------  dispose any leftover handles   --------------

	if (ph1 != nil)	{
	   #if defined(MACOS)			//	see also: BLYSTD.H for useful x-platform memory macros
		DisposeHandle(ph1);
	   #elif defined(WIN32OS)
		DeleteObject(ph1);			//	it's a bitmap object, not a handle, so don't use GlobalFree(ph1);
	   #endif
	}
	if (ph2 != nil)	{
	   #if defined(MACOS)
		DisposeHandle(ph2);
	   #elif defined(WIN32OS)
		DeleteObject(ph2);			//	it's a bitmap object, not a handle, so don't use GlobalFree(ph2);
	   #endif
	}

		//	--------------  return errors if any   --------------
		//	coerce to double and use SetFldVal, so negative numbers survive the trip

	if (errCd != noErr)	{
		xtnded = (odouble)errCd;
	   #if defined(powerc)
		//SetFldVal(ref_returnval, fmt_double, &xtnded, callback);	//	PPC version not working for this callback
		//SetFldLong(ref_returnval,ph1Size,callback);				//	or this one... (v3.6.4 callback file should fix it)
		NumToString((long)errCd,(unsigned char*)&strvalue[0]);		//	Meanwhile, convert to string and pass that back
		SetFldVal(ref_returnval, fmt_pstring, &strvalue[0], callback);	//	(problem is w/Metrowerks' support of long doubles)
	   #else
		SetFldVal(ref_returnval, fmt_double, &xtnded, callback);
	   #endif
	}

}	//	end of main

//	end of PictInfo 7CMD
//  IsPictureField  

Boolean IsPictureField( short fldRefNum, oproc callback)
{
	omnispbrec	fieldInfo;			//	parameter block record for GetFldInfo() call

	fieldInfo.name.len		= 0;
	fieldInfo.ref			= fldRefNum;
	fieldInfo.type			= 0;
	fieldInfo.info_request	= fir_field_info;
	if (GetFldInfo(&fieldInfo, callback) == woi_success)	{
		if (fieldInfo.type == fftPicture)	{
			return true;
		}
	}
	return false;
}

//	end of IsPictureField
//  DllMain  

/* Win32 DLL entry */

#if defined(WIN32OS)
BOOL WINAPI DllMain( HINSTANCE hInst, DWORD reason, LPVOID reserved)
{
#pragma unused(reserved)

	if (reason == DLL_PROCESS_ATTACH )	{
		if (hInstLib == 0)	{
			hInstLib = hInst;  					//	save instance handle for use in fetching strings
		}
	}
	else if (reason == DLL_PROCESS_DETACH)	{
		hInstLib = 0;
	}
	return 1;									//	return value ignored if reason != DLL_PROCESS_ATTACH
}
#endif	//	Win32 only

//	end of DllMain
//	  END  
